--##****************************************************
--##****************************************************
--##**	   TURRET DEFENCE		             **
--##**	              SCRIPT  BY SAYA                            **
--##**		    MAP BY VIEUXCHAT                    **
--##**	                   V 1.1 ALPHA                                       **
--##****************************************************



--scenario utilities
local ScenarioUtils = import('/lua/sim/ScenarioUtilities.lua')
local ScenarioFramework = import('/lua/ScenarioFramework.lua')
local Utilities = import('/lua/utilities.lua')
local EffectUtil = import('/lua/EffectUtilities.lua')

ScenarioInfo.Options.FogOfWar = 'none'



--PLAYERS, CONTAINS USEFUL INFO FOR THE SCRIPT FOR EACH PLAYER
local playersDATA = {
	['player1'] = {
		name = 'player1', 
		defeated = false,
		upgrader = false,
		ai = false,
		kills = 0,
		credits = 0,
		team = 1,
	},
	['player2'] = {
		name = 'player2', 
		defeated = false,
		upgrader = false,
		ai = false,
		kills = 0,
		credits = 0,
		team = 2,
	},
	['player3'] = {
		name = 'player3',
		defeated = false,
		upgrader = false,
		ai = false,
		kills = 0,
		credits = 0,
		team = 1,
	},
	['player4'] = {
		name = 'player4', 
		defeated = false,
		upgrader = false,
		ai = false,
		kills = 0,
		credits = 0,
		team = 2,
	},
	['player5'] = {
		name = 'player5', 
		defeated = false,
		upgrader = false,
		ai = false,
		kills = 0,
		credits = 0,
		team = 1,
	},
	['player6'] = {
		name = 'player6', 
		defeated = false,
		upgrader = false,
		ai = false,
		kills = 0,
		credits = 0,
		team = 2,
	},
	['player7'] = {
		name = 'player7',
		defeated = false,
		upgrader = false,
		ai = false,
		kills = 0,
		credits = 0,
		team = 1,
	},
	['player8'] = {
		name = 'player8', 
		defeated = false,
		upgrader = false,
		ai = false,
		kills = 0,
		credits = 0,
		team = 2,
	},
	['ai_team1'] = {
		name = 'ai_team1', 
		defeated = false,
		ai = true,
		color = {141,96,255},
		team = 1,
	},
	['ai_team2'] = {
		name = 'ai_team2',
		defeated = false,
		ai = true,
		color = {255,84,50},
		team = 2,
	},
	['neutral'] = {
		name = 'neutral',
		defeated = false,
		ai = true,
		color = {155,184,50},
	},
}

local teams = {
	['team1'] = {
		marker = 'team1',
		area = 'WIN_ZONE_TEAM1',
		defeated = false,
		name = 'team1',
	},
	['team2'] = {
		marker = 'team2',
		area = 'WIN_ZONE_TEAM2',
		defeated = false,
		name = 'team2',
	},
}



local players = {}

local armyTable = {
	 ['ARMY_1'] = 'player1',
	 ['ARMY_2'] = 'player2',
	 ['ARMY_3'] = 'player3',
	 ['ARMY_4'] = 'player4',
	 ['ARMY_5'] = 'player5',
	 ['ARMY_6'] = 'player6',
	 ['ARMY_7'] = 'player7',
	 ['ARMY_8'] = 'player8',
	 ['ARMY_9'] = 'ai_team1',
	 ['ARMY_10'] = 'ai_team2',
	 ['ARMY_11'] = 'neutral',
}


--scenario variables
local gameEnd = false
local executing = false
local beatTime = 2
local beat = 0


--map size
local xmapsize = ScenarioInfo.size[1]
local ymapsize = ScenarioInfo.size[2]
local mapHeight = 64
local mapRect = {x0 = 0, x1 = xmapsize, y0 = 0, y1 = ymapsize}


--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~GAME VARIABLES~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

--determines how much kills are needed to get a credit
local creditTimer = 30

--shop mass/credit cost tables
local packages = {
	--t1 packages
	{marker = "shop1", name = "20 Strikers", bpid = "", price = 1, composition = {{'UEL0201',20, wreckage = false},}, },
	{marker = "shop2", name = "35 Mech Marines", bpid = "", price = 1, composition = {{'UEL0106',35},}, },
	{marker = "shop3", name = "20 Mobile Anti-Air", bpid = "", price = 1, composition = {{'UEL0104',35},}, },
	{marker = "shop4", name = "15 Lobos", bpid = "", price = 1, composition = {{'UEL0103',15},}, },
	{marker = "shop5", name = "15 Bombers", bpid = "", price = 1, composition = {{'uea0103',15}} },
	{marker = "shop6", name = "20 Interceptors", bpid = "", price = 1, composition = {{'uea0102',20}} },
	{marker = "shop7", name = "5 Land Scouts, 5 Strikers, 5 Mech Marines, 5 Mobile AA", bpid = "", price = 1, composition = {{'UEL0101',5},{'UEL0201',5},{'UEL0106',5},{'UEL0104',2},} },
	{marker = "shop8", name = "10 Interceptors, 5 bombers, 5 Air Scouts", bpid = "", price = 1, composition = {{'uea0102',10},{'uea0103',5},{'uea0101',5}} },
	--special
	{marker = "shop9", name = "1 Runner", bpid = "", price = 4, composition = {{'url0303',1,health = 10000, speed = 5, name = 'Runner', effect = 1, weapons = false}}},
	--t2 packages
	--[[ Tier 2 Packages = Cost 3 Credits
	 1.20 Pillars
	 2.15 FlapJacks
	 3.5 Mobile Shields and 10 T2 Mobile AA
	 4.20 T2 Mobile AA
	 5.15 Gunships
	 6.10 Mobile Shields]]
	--t3 packages 
	--[[ Tier 3 Packages = Cost 8 Credits
	1.25 Titans
	2.15 Demolishers
	3.40 Wasps, 40 T1 Interceptors
	4.10 T3 Air Scouts, 30 Wasps, 20 T2 Mobile AA
	5.10 Strategic Bombers
	6.20 Broad Sword Gunships
	]]
	--special packages
	--[[Special Packages = Cost 12 Credits
	1.A Monkey Lord
	2.A CZAR

	Special Packages = Cost 20 Credits
	3.A Fatboy
	4.A Soul Ripper
	5.A Galactic Colossus

	Special Packages = Cost 40 Credits
	6.A Mavor
	]]
}



--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~MAIN SCRIPT FUNCTIONS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

import('/lua/SimUtils.lua').SetResourceSharing = function() end

--populating
function OnPopulate()	
	InitializeArmies()
    --ScenarioFramework.SetPlayableArea('PLAYABLE_AREA', false)	
end


--start of map script
function OnStart(self)

	-- force victory condition to sandbox
	--ScenarioInfo.Options.Victory = 'demoralization'
	ScenarioInfo.Options.Victory = 'sandbox'
	
	-- set score calculation to show player credits
	for index, brain in ArmyBrains do
	   brain.CalculateScore = function(thisBrain)
			local player = players[ArmyIndexer(thisBrain)]
			local score = 0
			
			if player.ai == false then
				score = players[ArmyIndexer(thisBrain)].credits
			end
			
			return score 
	   end
	end 
	
   
	--initialise scenario
	Utilities.UserConRequest("ui_ForceLifbarsOnEnemy")--show enemy life bars
	--allReclaimablesOFF()
	--initAreas()
	initShops()
	spawnStartUnits()
	 
	--start the different scenario threads
	startScenarioThreads()
end



--creates the different threads 
startScenarioThreads = function()

	--init thread
	local initThread = ForkThread(function(self)
		WaitSeconds(1)
		for index, player in players do 
			local brain = GetArmyBrain(toArmy(player))
			local mass = brain:GetEconomyStored('MASS')
			local startmoney = 500
			local energy = brain:GetEconomyStored('ENERGY')
			SetArmyEconomy( toArmy(player), -mass + startmoney , 0)		
			brain:GiveStorage('MASS', 2000)			
		end
			KillThread(self)
	end)
	
	--check for upgrade request thread
	local checkThread = ForkThread(function(self)
		while gameEnd == false do 
			WaitSeconds(1)
			checkShopRequests()
			checkVictoryConditions()
		end
		KillThread(self)
	end)
	
	--give credits to players regularly
	local creditThread = ForkThread(function(self)
		while gameEnd == false do
			WaitSeconds(creditTimer)
			for index,player in players do
				if player.ai == false and player.defeated == false then
					player.credits = player.credits + 1
				end
			end
		end
	end)
	

end



--check victory conditions 
checkVictoryConditions = function()
	--check all team victory areas
	for index,team in teams do
		local units = GetUnitsInRect(ScenarioUtils.AreaToRect(team.area))
		local enemyunits = {}
		if units and table.getn(units) >= 1 then
			for index,unit in  units do
				local teamname = 'team' ..  toPlayer(scnArmy(unit)).team
				if teamname != team.name then
					if not EntityCategoryContains(categories.AIR, unit) then
						--if enemy unit is in zone and it is not an air unit then defeat team concerned
						for index, player in players do
							if player.team then
								--defeat team and endgame
								local teamname = 'team' .. player.team
								team.defeated = true
								if player.team and teamname == team.name then
									player.defeated = true
									GetArmyBrain(toArmy(player)):OnDefeat()
								end
							end
						end
						
						--check if a team must win
						checkEndGame()
					end
				end
			end
		end
	end
end

--check if a team must win and make it win if necessary, compatible with more than 2 teams 
checkEndGame = function()
	--check for victors
	local teamsdefeated = 0
	local victoriousteam = false
	for index,team in teams do
		if team.defeated == true then
			teamsdefeated = teamsdefeated + 1
		else
			victoriousteam = team
		end
	end
	
	--if only one team has not been defeated it is the victorious team => game end
	if teamsdefeated >= table.getn(teams) - 1 then
		if victoriousteam != false then
			gameEnd = true
			for index, player in players do
				if player.ai == false then
					local teamname = 'team' .. player.team
					if player.team and teamname == victoriousteam.name then
						player.defeated = true
						GetArmyBrain(toArmy(player)):OnVictory()
					end
				end
			end
		end
	end
end

--data update for script ai
updateDATA = function()
	copyPlayerInfo()
end













--~~~~~~~~~~~~~~~~~~~~~~~~~~~~AREA MANAGEMENT~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

--given an area returns its center
areaCenter = function(area)
	local x = (area.x0 + area.x1)/2
	local y = (area.y0 + area.y1)/2
	return {x,y}
end

--given an area returns the central part of this area
processedArea = function(area)
	local areaWidth = math.abs(area.x1 - area.x0)
	local areaHeight = math.abs(area.y1 - area.y0)
	local center = areaCenter(area)
	local xa = center[1] - areaWidth/2
	local ya = center[2] - areaHeight/2
	local xb = center[1] + areaWidth/2
	local yb = center[2] + areaHeight/2
	return {x0 = xa, y0 = ya, x1 = xb, y1 = yb}	
end

















--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~PACKAGE MANAGEMENT~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

--spawn the current wave
spawnPackage = function(player,package)
	local aiBrain = GetArmyBrain(toArmy(player))
	local spawnArea = "WIN_ZONE_TEAM"..player.team
	local spawnRectangle = ScenarioUtils.AreaToRect(spawnArea)
	--local spawnRectangle = ScenarioUtils.AreaToRect("VALLEY")

	for index,unitset in package.composition do
		local units,data = spawnInRectangle(spawnRectangle,toArmy(player),unitset[1],5,unitset[2])
		
		-- WARN(table.getn(units))
		-- WARN(repr(data))
		
		for index, unit in units do
			doBuffs(unitset,unit)
		end
		
		local  relativerow = math.ceil(data.unitsSpawned/data.imax)*data.interval
		spawnRectangle.y0 = spawnRectangle.y0 + relativerow
		spawnRectangle.y1 = spawnRectangle.y1 + relativerow
	end
		
end

--do unit set buffs on a unit
--supports:
--health = int
--speed = int
--damage = int
--name = String
--effect = int
--wreckage = boolean
--slowdown = {int,int}
--poison = {int,int}
--invincible = boolean
--noweapons = boolean
--resistance = boolean
--income = int
--upgrades = String
doBuffs = function(unitset,unit)
	if unit and unit != false and not unit:IsDead() then
	
--play spawn effect 
if unitset.effect then 
   if unitset.effect == 1 then 
      spawnEffect(unit) 
   else 
   end 
end
		
		--specify speed
		if unitset.speed then
			unit:SetSpeedMult( unitset.speed )
			unit.originalSpeedmult = unitset.speed 
		end
		
		--specify health
		if unitset.health then	
			local health = unitset.health*(1 + (playerCount() - 1)*65/100)
			unit:SetMaxHealth(health)
			unit:SetHealth(nil,health)
		end
		
		--specify damage of unit weapons
		if unitset.damage then
			for i = 1, unit:GetWeaponCount() do
				setDamage(unit:GetWeapon(i),1)
			end
		end
		
		--disable damage call back (thus poison and slowdown) if true
		if unitset.resistance and unitset.resistance == true then
		else
			addDamagedCallBack(unit)
		end
		
		--disable wreckage if false
		if unitset.wreckage and unitset.wreckage == false then
			unit:GetBlueprint().Wreckage.WreckageLayers.Land=false
		end
		
		--disable weapons if false
		if unitset.weapons and unitset.weapons == false then
			unit:SetAllWeaponsEnabled(false)	
		end
		
		--give money for kills
		if unitset.income then
			addKillerIncome(unit,unitset.income,0)	
		end
		
		--apply upgrades if necessary
		if unitset.upgrades then 
			for i,enhancement in unitset.upgrades do
				unit:CreateEnhancement(enhancement)
			end
		end
		
		--slowdown ability
		if unitset.slowdown then
			addSlowDown(unit,unitset.slowdown[1],unitset.slowdown[2])
		end
		
		--poison ability
		if unitset.poison then
			addPoison(unit,unitset.poison[1],unitset.poison[2])
		end
		
		--turret name
		if unitset.name then
			unit:SetCustomName(unitset.name)
		end
		
		--unit is invincible
		if unitset.invincible and unitset.invincible == true then
			makeInvincible(unit)
		end
	end
end









--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~UNIT SPAWNING MANAGEMENT~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

--spawn the start buildings, life count buildings and shops
spawnStartUnits = function()
		
	for index, player in players do
		if player.ai == false then	
			respawnUpgrader(player)
		end
	end

end

--advanced function to spawn a certain number of a certain type of unit for a certain player inside a rectangle
--with a given interval and a maximum number of units
spawnInRectangle = function(rectangle,army,unitType,interval,maxunits)
	local units = {}
	
	local width = rectangle.x1 - rectangle.x0
	local height = rectangle.y1 - rectangle.y0
	
	local imax = math.floor(width/interval)
	local jmax = math.floor(height/interval)
	
	local unitsSpawned = 0
	
	--spawn units in rectangle within limit
	for j = 1,jmax do
		for i = 1,imax do				
			if unitsSpawned < maxunits then
				local x = rectangle.x0 + i*interval
				local y = rectangle.y0 + j*interval
				unit = CreateUnitHPR( unitType, army, x, mapHeight, y, 0,0,0) 
				--unit:GetBlueprint().Wreckage.WreckageLayers.Land=false
				table.insert(units,unit)
				unitsSpawned = unitsSpawned + 1
			end
		end
	end
	
	local data = {army = army,unitType = unitType,interval = interval,maxunits = maxunits,unitsSpawned = unitsSpawned,imax = imax,jmax = jmax,width = width,height = height, rectangle = rectangle}
	
	return (units or {}),data
end


--spawn unit at start position
spawnUnit = function(player,bpid)
	unit = CreateUnitHPR( bpid, toArmy(player),mPos(toArmy(player))[1], mPos(toArmy(player))[2],mPos(toArmy(player))[3], 0,0,0) 
	spawnEffect(unit)
	return unit
end

--spawn basic builder unit
createBuilder = function(player,bpid)
	if player.ai == false then
		local builder = false
		if bpid then 
			builder = spawnUnit(player,bpid)	
		else
			builder = spawnUnit(player,'uel0301')		
		end		
		makeBuilder(player,builder)
	end
end













--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~SHOP AND REQUEST MANAGEMENT FUNCTIONS~~~~~~~~~~~~~~~~~~~~~~

--~~~~Particular shop behaviour~~~~

--list of all shops with their callback functions
local shops = {}

--create one shop per package
createPackageShops = function()
	for k,v in packages do
		local package = table.deepcopy(v)
		for index, team in teams do 
			if hasMembers(team) then
				local shop = {name = team.marker .. '_' .. package.marker, description = package.name .. ' : ' .. package.price .. ' credits', bpid = package.bpid, team = team.name, buy = function(player) buyPackage(player,package) end}
				table.insert(shops,shop)
			end
		end
	end
end

--buy a package, spawn it if you have enough credits
buyPackage = function(player,package)
	if checkPrice(player,package.price,"credits") then 
		spawnPackage(player,package)
		fastPrint(username(player) .. ' bought ' .. package.name) 
	else 
		fastFocusPrint('Not enough credits for ' .. package.name,player)
	end
end

--~~~~Global Shop Management~~~~

--initialise all shops
initShops = function()
	createPackageShops()
	
	for index, shop in shops do
		if GetMarker(shop.name) then
			local pos = mPos(shop.name)
			local army = toArmy(players['neutral'])
			
			if shop.team == 'team1' then
				army = toArmy(players['ai_team1'])
			elseif shop.team == 'team2' then
				army = toArmy(players['ai_team2'])
			end
			
			--use unit type or default
			if shop.bpid and shop.bpid != "" then
				shop.unit = CreateUnitHPR( shop.bpid, army, pos[1], pos[2], pos[3], 0,0,0) 
			else
				shop.unit = CreateUnitHPR( 'ueb5202', army, pos[1], pos[2], pos[3], 0,0,0) 
			end
			
			--use shop description or default
			if shop.description then
				shop.unit:SetCustomName(shop.description)
			else
				shop.unit:SetCustomName(shop.name)
			end
			
			--make shope invincible
			shop.unit:SetAllWeaponsEnabled(false)		
			shop.unit:SetCanBeKilled(false)
			shop.unit:SetDoNotTarget(true)
			shop.unit:SetMaxHealth(1)
			shop.unit:SetHealth(nil,1)
			shop.unit:SetRegenRate(1)
			shop.unit:SetUnSelectable(true)
			shop.unit:SetImmobile(true)
		else
			shop.unit = false
		end
	end
end

--checks all shops for each players
checkShopRequests = function()
	for index, shop in shops do 
		checkShop(shop)
	end
end

--check if there is a shop request and do the corresponding action if the check is positive
checkShop = function(shop)
	if shop.unit != false then	
		for index,player in players do
			if player.upgrader then
				--get target data
				local v1 = player.upgrader:GetNavigator():GetGoalPos()
				local v2 = shop.unit:GetPosition()
				local check = {false,false,false}
				local tolerance = 10
				
				--verify if target is inside range
				for i = 1,3 do
					if math.abs(v1[i] - v2[i]) <= tolerance then
						check[i] = true
					end
				end

				--if check is ok do shop action
				if check[1] == true and check[3] == true then --and check[2] == true --ignoring vertical checking			   
					ForkThread(function() playBuildEffect(shop.unit) end)					
					respawnUpgrader(player)
					shop.buy(player)
				end
			end
		end
	end
end


--kill and respawn player upgrader  unit
respawnUpgrader = function(player)
	
	--calculate spawn position
	local teamNumber = player.team
	local position = mPos('team' .. teamNumber .. '_upgrader') 
	
	local xPos = position[1]
	local zPos = position[2]
	local yPos = position[3]

	--destroy upgrader unit if necessary
	if player.upgrader != false then
		player.upgrader:SetCanBeKilled(true)
		player.upgrader:Kill()
		player.upgrader = false
	end	
	--spawn upgrader unit
	if player.upgrader == false then 
		player.upgrader = CreateUnitHPR( 'uec0001', toArmy(player), xPos,zPos,yPos, 0,0,0)
		player.upgrader:GetBlueprint().Wreckage.WreckageLayers.Land=false		
		player.upgrader:SetAllWeaponsEnabled(false)	
		player.upgrader:SetCanBeKilled(false)
		player.upgrader:SetDoNotTarget(true)
		player.upgrader:SetMaxHealth(1)
		player.upgrader:SetHealth(nil,1)
		player.upgrader:SetRegenRate(1)
		player.upgrader:SetSpeedMult(2)
		player.upgrader:SetAccMult(10)       
        --player.upgrader:SetTurnMult(10)
		player.upgrader:SetCustomName(username(player) .. ': business man')
	end
end







--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~UNIT BEHAVIOUR CHANGE FUNCTIONS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


--makes a unit invincible
makeInvincible = function(unit)
	unit:SetDoNotTarget(true)
	unit:SetCanBeKilled(false)
	unit:SetCapturable(false)
	unit:SetReclaimable(false)
	unit:SetRegenRate(1)
	unit:SetMaxHealth(1)
	unit:SetHealth(nil,1)
end

--unit killer gets income
addKillerIncome = function(unit,massincome,energyincome)
	unit.OldOnKilled = unit.OnKilled
	unit.OnKilled = function(self, instigator, type, overkillRatio)
		--give 5 mass income per kill
		if instigator then
			local player = players[armyTable[scnArmy(self)]]
			local killer = players[armyTable[scnArmy(instigator)]]
			SetArmyEconomy( toArmy(killer), massincome  , energyincome)	
			killer.kills = killer.kills + 1
			--checkCredits(killer)
		end
		unit.OldOnKilled(self, instigator, type, overkillRatio)	
	end	
end

--give a credit to the player if necessary
checkCredits = function(player)
	local modulo = math.mod(player.kills,creditKillCost)
	if modulo == 0 then
		player.credits = player.credits + 1
	end
end

--unit hp will be set to lifetime and unit will lose one hp per second until it dies
makePerishable = function(unit,lifetime)
	--unit.OnDamage = function(self) WARN("LOL") end
	unit:SetMaxHealth(lifetime)
	local timeleft = lifetime
	local lifetimeThread = ForkThread(function(self)	
		for timeleft = lifetime,0,-1 do
			if unit and not unit:IsDead() then
				unit:SetHealth(nil,timeleft)
			end
			WaitSeconds(1)
		end
		if unit and not unit:IsDead() then
			unit:Kill()
		end
		KillThread(self)
	end)
end


--replaces the normal onfire sequence by a custom weapon firing pattern
--weapon hits instantly every time it shoots.
addInstaHit = function(instigator)
	local bp = instigator:GetBlueprint()
	local bpid = bp.BlueprintId
	local turret = getTurret(bpid)
	
	for i = 1, instigator:GetWeaponCount() do
		if not turret.weaponsfired or (turret.weaponsfired and i <= turret.weaponsfired) then	
			local weapon = instigator:GetWeapon(i)
			weapon.OnFire = function(weapon)	

				--play weapon sound
				local wbp = weapon:GetBlueprint()
		        if wbp.Audio.Fire then
		            weapon:PlaySound(wbp.Audio.Fire)
		        end
			
				--aquire target and convert to unit if target is a blip
				local target = weapon:GetCurrentTarget()
				if IsBlip(target) then target = target:GetSource() end
				
				if turret.area and turret.area != 0 then
				
					--if area damage, get units around point
					if GetArmyBrain(scnArmy(target)):GetUnitsAroundPoint(categories.MOBILE,target:GetPosition(), turret.area,'Ally') then
						local targetUnits = GetArmyBrain(scnArmy(target)):GetUnitsAroundPoint(categories.MOBILE,target:GetPosition(), turret.area,'Ally')
						if table.getsize(targetUnits) > 0 then
							for index,target in targetUnits do
								if turret.damage > 0 then
									--do damage
									Damage(instigator, target:GetPosition(), target, turret.damage, 'Normal')
								end
								--do custom unit abilities
								doAbilities(target,instigator)
							end
						end
					end
					
				else		
					if turret.damage > 0 then
						--do damage
						Damage(instigator, target:GetPosition(), target, turret.damage, 'Normal')
					end
					--do custom unit abilities
					doAbilities(target,instigator)
				end

				--weapon.OnLostTarget(weapon)
				

				
			end
		else
			weapon.OnFire = function(weapon) end
		end
	end
end

--do custom unit abilities such as freeze and poison
doAbilities = function(target,instigator)
	if target and instigator and not target:IsDead() and target.HasDamageCallbacks and target.HasDamageCallbacks == true then
		if instigator.slowdown then
			doSlowDown(target,instigator,instigator.slowdown.percent,instigator.slowdown.time)
		end
		if instigator.poison then
			doPoison(target,instigator.poison)
		end
	end
end

--add an ondamaged callback
addDamagedCallBack = function(unit)
	unit.HasDamageCallbacks = true 
end


--ad poisoning ability to unit
addPoison = function(unit,poisondamage,poisontime)
	unit.poison = {damage = poisondamage, time = poisontime}
	addInstaHit(unit)
end

--add slowdown ability to unit
addSlowDown = function(unit,slowdownpercent,slowtime)
	unit.slowdown = {percent = slowdownpercent, time = slowtime}
	addInstaHit(unit)
end



--poison a unit using a poison table containing damage persecond and poisoning time
poison = function(unit,poison,effects)	
	ForkThread(function()
		for timeleft = poison.time,0,-1 do
			WaitSeconds(1)
			if unit and not unit:IsDead() then
				if unit:GetHealth() > poison.damage then
					unit:AdjustHealth(nil, -poison.damage)
				else
					unit:SetHealth(nil,1)
				end
			end
		end
		if effects then
			for k,effect in effects do
				effect:Destroy()
			end
		end
	end)
end

--poison a unit with effects
doPoison = function(unit,poisontable)	
	local emittereffect = '/effects/emitters/aeon_chrono_dampener_04_emit.bp'
	local emitter = CreateAttachedEmitter(unit, unit:GetBoneName(0), unit:GetArmy(), emittereffect)	
	local effects = {emitter}
	
	poison(unit,poisontable,effects)
end

--slow down freezes a unit for a certain number of seconds by a certain percent of its speed. 
slowDown = function(unit,slowdownpercent,slowtime,effects)
	local slowdown
	if unit.originalSpeedmult then
		slowdown = unit.originalSpeedmult*(slowdownpercent/100)
	else
		slowdown = (slowdownpercent/100)
	end
	
	unit:SetSpeedMult(slowdown)
	ForkThread(function() 
		WaitSeconds(slowtime)	
		if unit and not unit:IsDead() then
			if effects then
				for k,effect in effects do
					effect:Destroy()
				end
			end
			if unit.originalSpeedmult then
				unit:SetSpeedMult(unit.originalSpeedmult)
			else
				unit:SetSpeedMult(1)
			end
		end
	end)
end


--slow down unit with beam effect
doSlowDown = function(unit,instigator,slowdownpercent,slowtime)
	local beameffect = '/effects/emitters/terran_transport_beam_01_emit.bp'
	local beameffect2 = '/effects/emitters/terran_transport_beam_02_emit.bp'
	--local emittereffect = '/effects/emitters/terran_transport_glow_01_emit.bp'

	local beam = AttachBeamEntityToEntity(unit, unit:GetBoneName(0), instigator, instigator:GetBoneName(0), instigator:GetArmy(), beameffect)
	local beam2 = AttachBeamEntityToEntity(unit, unit:GetBoneName(0), instigator, instigator:GetBoneName(0), instigator:GetArmy(), beameffect2)
	--local emitter = CreateEmitterAtBone( instigator, instigator:GetBoneName(0), instigator:GetArmy(), emittereffect)
	
	local effects = {beam,beam2}

	slowDown(unit,slowdownpercent,slowtime,effects)
end

--play a nice build effect
playBuildEffect = function(unit)
	unit:PlayUnitSound('TeleportStart')
	unit:PlayUnitAmbientSound('TeleportLoop')
	WaitSeconds( 0.1 )
	unit:PlayTeleportInEffects()
	WaitSeconds( 0.1 ) 
    unit:StopUnitAmbientSound('TeleportLoop')
    unit:PlayUnitSound('TeleportEnd')
end














--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~MISCALENIOUS FUNCTIONS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~




--SPECIAL ARMY INITIALISATION: CUSTOM ALLIANCES
function InitializeArmies()
    local tblGroups = {}
    local tblArmy = ListArmies()
	
	--determine the players depending on which slots are occupied
	for index,army in tblArmy do
		players[armyTable[army]] = playersDATA[armyTable[army]]
		ScenarioFramework.AddRestriction(army, categories.WALL)
	end
	
	--initialise armies
    for iArmy, strArmy in pairs(tblArmy) do
        local tblData = Scenario.Armies[strArmy]

        tblGroups[ strArmy ] = {}

        if tblData then
			
			AddBuildRestriction(iArmy,buildRestriction)
			--SetArmyEconomy( strArmy, tblData.Economy.mass, tblData.Economy.energy)

        end

    end

	--my function to initialise alliances
	initialiseAlliances()
	
	--GetArmyBrain('ARMY_9').civilian = false
	
    return tblGroups
end



--the defender is alone against the attackers
function initialiseAlliances()
	for index,army in ListArmies() do

		--a player is allied to everybody exept its direct enemy
		for index2,army2 in ListArmies() do
			--set ai colors
			if toPlayer(army).ai == true then
				SetArmyColor(armyToIndex(army), toPlayer(army).color[1], toPlayer(army).color[2], toPlayer(army).color[3])			
			end
			--set alliances
			if toPlayer(army).team  and toPlayer(army2).team then
				if toPlayer(army2).team == toPlayer(army).team then
					SetAlliance( army, army2, 'Ally')
				else
					SetAlliance( army, army2, 'Enemy')
				end
			end
		end

	end
end

















--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~UTILITY FUNCTIONS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

--given a player returns an army
toArmy = function(player)
	local army = false
	for i,v in armyTable do
		if v == player.name then
			army = i
		end
	end
	return army
end


--given a player returns a proper username
username = function(player)
	return GetArmyBrain(toArmy(player)).Nickname
end

--given a player returns the number of kills
armyKills = function(player)
	local kills = 0
	if player.name != 'neutral' then 
		local army = toArmy(player)
		kills = GetArmyBrain(army):GetArmyStat('Enemies_Killed',0.0).Value 
	end
	return kills	
end

--print a text scn way :p
--first arg is the text to print
--second arg is what player it concerns
--third arg is if it has to be printed to that player only or to all players
scnPrint = function(text,player,playeronly)

	local army = toArmy(player)
	local colorIndex = ScenarioInfo.ArmySetup[army].PlayerColor
	local color = import('/lua/GameColors.lua').GameColors.ArmyColors[colorIndex]
	
	if not playeronly then
		playerspecific = false
	else
		playerspecific = playeronly
	end
	
	
	if playerspecific == true then
		if army == indexToArmy(GetFocusArmy()) then
			PrintText(text,20,color,5,'center') 
		end
	else
		PrintText(text,20,color,5,'center') 
	end
end

fastFocusPrint = function(text,player)
	local army = toArmy(player)
	if army == indexToArmy(GetFocusArmy()) then
		PrintText(text,20,nil,5,'center') 	
	end	
end

fastPrint = function(text)
	PrintText(text,20,nil,5,'center') 		
end

--given a unit returns the army
scnArmy = function(unit)
	local armyIndex = unit:GetArmy()
	return indexToArmy(armyIndex)
end

--given an army index returns an army
indexToArmy = function(armyIndex)
	local army = ListArmies()[armyIndex]
	return army
end

--given an army return its index
armyToIndex = function(army)
	for i,v in ListArmies() do
		if v == army then
			return i 
		end
	end
end

--get the number of players
playerCount = function()
	local i = 0
	for index, player in players do
		if player.ai == false then
			i = i + 1
		end
	end
	return i
end


--given an ai brain returns an army index
function ArmyIndexer(aiBrain)
	local tblArmy = ListArmies()
	local army = tblArmy[aiBrain:GetArmyIndex()]
	for i, v in armyTable do
		if i == army then
			return v
		end
	end
end

--given an army returns a player
function toPlayer(army)
	return players[armyTable[army]]
end

--given a team number returns its neutral team player
teamToArmy = function(team)
	local army = 'ARMY_'
	if team == 1 then
		army = army .. '9'
	elseif team == 2 then
		army = army .. '10' 
	end
	return army
end

teamName = function(player)
	return 'team' .. player.team
end

hasMembers = function(team)
	local members = false
	for index,player in players do
		if player.ai == false then
			if teamName(player) == team.name then
				members = true	
			end
		end
	end
	return members
end

--copy the map script variables into the aibrains so that the zone control ai can use them
copyPlayerInfo = function()
	for i, player in players do
		GetArmyBrain(toArmy(player)).SAODnfo = table.copy(players)
	end
end

--get marker given a name from map_save done in map editor, you can the use marker.position to get a location in the {x,y,z} format
function GetMarker(name)
    return Scenario.MasterChain._MASTERCHAIN_.Markers[name]
end

function mPos(name)
	return GetMarker(name).position
end

--given a marker return a rectangle
markerToRect = function(markername)
	local rectangle = {}
	local position = GetMarker(markername).position

	if not position then position = {0,0,0} end
	
	rectangle.x0 = position[1] - 25
	rectangle.y0 = position[3] - 25
	rectangle.x1 = position[1] + 25
	rectangle.y1 = position[3] + 25
	
	--special treatment for start and finish
	if markername == "Start" then
		rectangle.startarea = true
	elseif markername == "Finish" then
		rectangle.finalarea = true
	end
	
	return rectangle
end

--display a countdown
--call from within a thread
countDown = function(timeseconds)
	for timeleft = timeseconds,0,-1 do 
		Sync.ObjectiveTimer = timeleft
		WaitSeconds(1)
	end
end

--check if a player has enough money and if he has takes that money from him and returns true
--mass is used as money here
checkPrice = function(player,price,pricetype)
	local army = toArmy(player)
	local brain = GetArmyBrain(army)
			
	if pricetype then
		if pricetype == "credits" then
			if checkCreditPrice(player,price) then return true else return false end
		end
	else
		
		if brain:GetEconomyStored('MASS') >= price then	
			SetArmyEconomy(army,-price,0 )	
			return true
		else
			return false
		end
		
	end
end

checkCreditPrice = function(player,price)
	if player.credits >= price then	
		player.credits = player.credits - price
		return true
	else
		return false
	end
end

--play a nice effect on a unit
spawnEffect = function(unit)
	unit:PlayTeleportInEffects()
end



--all rock and tress on the map will not be reclaimable
allReclaimablesOFF = function()
	entities = GetEntitiesInRect(mapRect)
	reclaimables = {}
	for i,ent in entities do
		if (ent.MassReclaim and ent.MassReclaim > 0) or (ent.EnergyReclaim and ent.EnergyReclaim > 0) then
			table.insert(reclaimables,ent)
		end
	end
	if reclaimables and table.getn(reclaimables) then
		for i,ent in reclaimables do
			ent.MassReclaim = 0
			ent.EnergyReclaim = 0
		end
	end
end

--simple function to set the damage of a weapon
setDamage = function(weapon,value)
	local bp = weapon:GetBlueprint()
	weapon.DamageMod = 0
	weapon:AddDamageMod(-bp.Damage)
	weapon:AddDamageMod(value)
end

